import logging
import os
import unittest
from quantastica.qiskit_toaster import ToasterBackend
from qiskit import QuantumRegister, ClassicalRegister
from qiskit import QuantumCircuit, execute, Aer
from qiskit.compiler import transpile, assemble
from math import pi

class TestToasterBackend(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(
            format='%(levelname)s %(asctime)s %(pathname)s - %(message)s',
            level=os.environ.get("LOGLEVEL", "CRITICAL"),
        )
        pass

    def tearDown(self):
        pass

    def test_bell_counts(self):
        shots = 256
        qc = TestToasterBackend.get_bell_qc()
        stats = TestToasterBackend.execute_and_get_stats(
            ToasterBackend.ToasterBackend(), qc, shots
        )
        self.assertTrue(stats["statevector"] is None)
        self.assertEqual(len(stats["counts"]), 2)
        self.assertEqual(stats["totalcounts"], shots)

    def test_bell_counts_with_seed(self):
        shots = 1024
        qc=TestToasterBackend.get_bell_qc()
        stats1 = TestToasterBackend.execute_and_get_stats(
            ToasterBackend.ToasterBackend(),
            qc,
            shots,
            seed = 1
        )
        stats2 = TestToasterBackend.execute_and_get_stats(
            ToasterBackend.ToasterBackend(),
            qc,
            shots,
            seed = 1
        )
        stats3 = TestToasterBackend.execute_and_get_stats(
            ToasterBackend.ToasterBackend(),
            qc,
            shots,
            seed = 2
        )
        self.assertTrue( stats1['statevector'] is None)
        self.assertEqual( len(stats1['counts']), 2)
        self.assertEqual( stats1['totalcounts'], shots)
        self.assertEqual(stats1['counts'],stats2['counts'])
        self.assertNotEqual(stats1['counts'],stats3['counts'])

    def test_teleport_counts(self):
        shots = 256
        qc = TestToasterBackend.get_teleport_qc()
        stats = TestToasterBackend.execute_and_get_stats(
            ToasterBackend.ToasterBackend(), qc, shots
        )
        self.assertTrue(stats["statevector"] is None)
        self.assertEqual(stats["totalcounts"], shots)
        self.assertEqual(len(stats["counts"]), 4)

    def test_bell_state_vector(self):
        """
        This is test for statevector which means that
        even with shots > 1 it should execute only one shot
        """
        shots = 256
        qc = TestToasterBackend.get_bell_qc()
        stats = TestToasterBackend.execute_and_get_stats(
            ToasterBackend.ToasterBackend(
                backend_name="statevector_simulator"
            ),
            qc,
            shots,
        )
        self.assertEqual(len(stats["statevector"]), 4)
        self.assertEqual(len(stats["counts"]), 1)
        self.assertEqual(stats["totalcounts"], 1)

    def test_teleport_state_vector(self):
        """
        This is test for statevector which means that
        even with shots > 1 it should execute only one shot
        """
        shots = 256
        qc = TestToasterBackend.get_teleport_qc()

        """
        Let's first run the aer simulation to get statevector
        and counts so we can compare those results against forest's
        """
        stats_aer = TestToasterBackend.execute_and_get_stats(
            Aer.get_backend("statevector_simulator"), qc, shots
        )
        """
        Now execute forest backend
        """
        stats = TestToasterBackend.execute_and_get_stats(
            ToasterBackend.ToasterBackend(
                backend_name="statevector_simulator"
            ),
            qc,
            shots,
        )

        self.assertEqual(len(stats["counts"]), len(stats_aer["counts"]))
        self.assertEqual(
            len(stats["statevector"]), len(stats_aer["statevector"])
        )
        self.assertEqual(stats["totalcounts"], stats_aer["totalcounts"])

        """
        Let's verify that tests are working as expected
        by running fail case
        """
        stats = TestToasterBackend.execute_and_get_stats(
            ToasterBackend.ToasterBackend(), qc, shots
        )
        self.assertTrue(stats["statevector"] is None)
        self.assertNotEqual(stats["totalcounts"], stats_aer["totalcounts"])

    def test_multiple_jobs(self):
        qc = self.get_bell_qc()
        backend = ToasterBackend.ToasterBackend()
        jobs = []
        for i in range(1, 50):
            jobs.append(execute(qc, backend=backend, shots=1))
        for job in jobs:
            result = job.result()
            counts = result.get_counts(qc)
            self.assertEqual(len(counts), 1)

    def test_multiple_experiments(self):
        backend = ToasterBackend.ToasterBackend()
        qc_list = [ self.get_bell_qc(), self.get_teleport_qc() ]
        transpiled = transpile(qc_list, backend = backend)
        qobjs = assemble(transpiled, backend=backend, shots=4096)
        job_info = backend.run(qobjs)
        bell_counts = job_info.result().get_counts("Bell")
        tel_counts = job_info.result().get_counts("Teleport")
        self.assertEqual(len(bell_counts),2)
        self.assertEqual(len(tel_counts),4)

    @staticmethod
    def execute_and_get_stats(backend, qc, shots, seed = None):
        job = execute(qc, backend=backend, shots=shots, seed_simulator = seed)
        job_result = job.result()
        counts = job_result.get_counts(qc)
        total_counts = 0
        for c in counts:
            total_counts += counts[c]

        try:
            state_vector = job_result.get_statevector(qc)
        except:
            state_vector = None
        ret = dict()
        ret["counts"] = counts
        ret["statevector"] = state_vector
        ret["totalcounts"] = total_counts
        return ret

    @staticmethod
    def get_bell_qc():
        qc = QuantumCircuit(name="Bell")

        q = QuantumRegister(2, "q")
        c = ClassicalRegister(2, "c")

        qc.add_register(q)
        qc.add_register(c)

        qc.h(q[0])
        qc.cx(q[0], q[1])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[1])
        return qc

    @staticmethod
    def get_teleport_qc():
        qc = QuantumCircuit(name="Teleport")

        q = QuantumRegister(3, "q")
        c0 = ClassicalRegister(1, "c0")
        c1 = ClassicalRegister(1, "c1")

        qc.add_register(q)
        qc.add_register(c0)
        qc.add_register(c1)

        qc.rx(pi / 4, q[0])
        qc.h(q[1])
        qc.cx(q[1], q[2])
        qc.cx(q[0], q[1])
        qc.h(q[0])
        qc.measure(q[1], c1[0])
        qc.x(q[2]).c_if(c1, 1)
        qc.measure(q[0], c0[0])
        qc.z(q[2]).c_if(c0, 1)
        return qc


if __name__ == "__main__":
    unittest.main()



# This code is part of quantastica.qiskit_toaster
#
# (C) Copyright Quantastica 2019.
# https://quantastica.com/
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from concurrent import futures
import logging
import subprocess
import json
import time
import copy

from quantastica.qconvert import qobj_to_toaster

from qiskit.providers import BaseJob, JobStatus, JobError
from qiskit.result import Result

logger = logging.getLogger(__name__)

def _run_with_qtoaster_static(qobj_dict, get_states, toaster_path, job_id):
    SEED_SIMULATOR_KEY = "seed_simulator"
    if get_states :
        shots = 1
    else:
        shots = qobj_dict['config']['shots']
    args = [
        toaster_path,
        "-",
        "-r",
        "counts",
        "-s",
        "%d"%shots
    ]
    if get_states:
        args.append("-r")
        args.append("state")
    seed = 0
    if SEED_SIMULATOR_KEY in qobj_dict['config']:
        seed = qobj_dict['config'][SEED_SIMULATOR_KEY]
        args.append("--seed")
        args.append("%d"%seed)
    logger.info(args)
    proc = subprocess.Popen(
        args,
        close_fds = False,
        restore_signals = False,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE )

    converted = qobj_to_toaster(qobj_dict, { "all_experiments": False })

    logger.info("Running q-toaster with following params:")
    proc.stdin.write(converted.encode())
    proc.stdin.close()
    proc.wait()
    qtoasterjson = proc.stdout.read()
    proc.stdout.close()
    resultraw = json.loads(qtoasterjson)
    logger.debug("Raw results from toaster:\r\n%s",resultraw)
    if isinstance(resultraw , dict) :
        rawversion = resultraw.get('qtoaster_version')
    else:
        rawversion = "0.0.0"

    if ToasterJob._check_qtoaster_version(rawversion) == False :
        raise ValueError(
            "Unsupported qtoaster_version, got '%s' - minimum expected is '%s'.\n\rPlease update your q-toaster to latest version"%(rawversion,ToasterJob._MINQTOASTERVERSION))

    counts = ToasterJob._convert_counts(resultraw['counts'])
    exp_dict = qobj_dict['experiments'][0]
    exp_header = exp_dict['header']
    expname = exp_header['name']
    statevector = resultraw.get('statevector')
    data = dict()
    data['counts'] = counts;
    if statevector is not None and len(statevector) > 0:
        data['statevector'] = statevector

    result = {
                'success': True,
                'meas_level': 2,
                'shots': shots,
                'data': data,
                'header': exp_header,
                'status': 'DONE',
                'time_taken': resultraw['time_taken'],
                'name': expname,
                'seed_simulator': seed,
                'toaster_version' : rawversion
            }
    return result

class ToasterJob(BaseJob):
    _MINQTOASTERVERSION = '0.9.9'

    _executor = futures.ProcessPoolExecutor()
    _run_time = 0

    def __init__(self, backend, job_id, qobj, toasterpath,
                 getstates = False):
        super().__init__(backend, job_id)
        self._result = None
        self._qobj_dict = qobj.to_dict()
        self._futures = []
        self._toasterpath = toasterpath
        self._getstates = getstates

    def submit(self):
        if len(self._futures)>0:
            raise JobError("We have already submitted the job!")
        self._t_submit = time.time()

        logger.debug("submitting...")
        all_exps = self._qobj_dict
        for exp in all_exps["experiments"]:
            single_exp = copy.deepcopy(all_exps)
            single_exp["experiments"]=[exp]
            self._futures.append(self._executor.submit(_run_with_qtoaster_static,
                single_exp,
                self._getstates,
                self._toasterpath,
                self._job_id)
                )

    def wait(self, timeout=None):
        if self.status() in [JobStatus.RUNNING, JobStatus.QUEUED] :
            futures.wait(self._futures, timeout)
        if self._result is None and self.status() is JobStatus.DONE :
            results = []
            for f in self._futures:
                results.append(f.result())
            qobj_dict = self._qobj_dict
            qobjid = qobj_dict['qobj_id']
            qobj_header = qobj_dict['header']
            rawversion = "1.0.0"
            if len(results):
                if "toaster_version" in results[0]:
                    rawversion = results[0]['toaster_version']

            self._result = {
                'success': True,
                'backend_name': "Toaster",
                'qobj_id': qobjid ,
                'backend_version': rawversion,
                'header': qobj_header,
                'job_id': self._job_id,
                'results': results,
                'status': 'COMPLETED'
            }
            ToasterJob._run_time += time.time() - self._t_submit

        if len(self._futures)>0:
            for f in self._futures:
                if f.exception() :
                    raise f.exception()

    def result(self, timeout=None):
        self.wait(timeout)
        return Result.from_dict(self._result)

    def cancel(self):
        return
        #return self._future.cancel()

    def status(self):

        if len(self._futures)==0 :
            _status = JobStatus.INITIALIZING
        else :
            running = 0
            done = 0
            canceled = 0
            error = 0
            queued = 0
            for f in self._futures:
                if f.running():
                    running += 1
                elif f.cancelled():
                    canceled += 1
                elif f.done():
                    if f.exception() is None:
                        done += 1
                    else:
                        error += 1
                else:
                    queued += 1

            if error :
                _status = JobStatus.ERROR
            elif running :
                _status = JobStatus.RUNNING
            elif canceled :
                _status = JobStatus.CANCELLED
            elif done :
                _status = JobStatus.DONE
            else: # future is in pending state
                _status = JobStatus.QUEUED
        return _status

    def backend(self):
        """Return the instance of the backend used for this job."""
        return self._backend

    @staticmethod
    def _qtoaster_version_to_int(versionstring):
        parts = versionstring.split(".")
        verint = 0
        if len(parts) >= 3 :
            verint = int(parts[0])*1000000 + int(parts[1])*1000 + int(parts[2])
        return verint

    @classmethod
    def _check_qtoaster_version(cls, versionstring):
        verint = cls._qtoaster_version_to_int(versionstring)
        minverint = cls._qtoaster_version_to_int(cls._MINQTOASTERVERSION)
        if verint < minverint :
            return False
        return True

    @staticmethod
    def _convert_counts(counts):
        ret = dict()
        for key in counts:
            nicekey = key.replace(' ','')
            nicekey = hex(int(nicekey,2))
            ret[nicekey] = counts[key]
        return ret



import unittest
import networkx as nx
import numpy as np
from docplex.mp.model import Model

from qiskit import BasicAer
from qiskit.aqua import aqua_globals, QuantumInstance
from qiskit.aqua.algorithms import QAOA
from qiskit.aqua.components.optimizers import SPSA
from qiskit.optimization.ising import docplex, max_cut
from qiskit.optimization.ising.common import sample_most_likely
from quantastica.qiskit_toaster import ToasterBackend, ToasterJob


import time
import sys
import logging
import os


@unittest.skipUnless(
    os.getenv("SLOW") == "1",
    "Skipping this test (environment variable SLOW must be set to 1)",
)
class TestQAOA(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(
            format='%(levelname)s %(asctime)s %(pathname)s - %(message)s',
            level=os.environ.get("LOGLEVEL", "CRITICAL"),
        )
        self.startTime = time.time()

    def tearDown(self):
        t = time.time() - self.startTime
        sys.stderr.write(" took %.3fs ... " % (t))

    def test_qaoa(self):
        print("Running toaster test...")
        toaster_backend = ToasterBackend.get_backend("qasm_simulator")
        toaster_results = self.run_simulation(toaster_backend)
        print("Run time:",ToasterJob.ToasterJob._run_time,"seconds")
        print("Running AER test...")
        aer_backend = BasicAer.get_backend("qasm_simulator")
        aer_results = self.run_simulation(aer_backend)
        print("===== Calculations done =====")
        print("  ==== AER Results =====")
        print(aer_results)
        print("  ==== Toaster Results =====")
        print(toaster_results)
        threshold = 0.9
        aer_k = abs(aer_results['maxcut_objective']/aer_results['solution_objective'])
        toaster_k = abs(toaster_results['maxcut_objective']/toaster_results['solution_objective'])
        self.assertGreater( aer_k, threshold )
        self.assertGreater( toaster_k, threshold )

    def run_simulation(self, backend):
        #
        # Random 3-regular graph with 12 nodes
        #
        n      = int(os.environ.get("N", "4"))
        graph = nx.random_regular_graph(3, n)
        for e in graph.edges():
            graph[e[0]][e[1]]['weight'] = 1.0


        # Compute the weight matrix from the graph
        w = np.zeros([n, n])
        for i in range(n):
            for j in range(n):
                temp = graph.get_edge_data(i, j, default=0)
                if temp != 0:
                    w[i, j] = temp["weight"]

        # Create an Ising Hamiltonian with docplex.
        mdl = Model(name="max_cut")
        mdl.node_vars = mdl.binary_var_list(list(range(n)), name="node")
        maxcut_func = mdl.sum(
            w[i, j] * mdl.node_vars[i] * (1 - mdl.node_vars[j])
            for i in range(n)
            for j in range(n)
        )
        mdl.maximize(maxcut_func)
        qubit_op, offset = docplex.get_operator(mdl)

        # Run quantum algorithm QAOA on qasm simulator
        seed = int(os.environ.get("SEED", "40598"))
        aqua_globals.random_seed = seed

        spsa = SPSA(max_trials=250)
        qaoa = QAOA(qubit_op, spsa, p=5, max_evals_grouped = 4)

        quantum_instance = QuantumInstance(
            backend, shots=1024, seed_simulator=seed, seed_transpiler=seed,
            optimization_level=0
        )
        result = qaoa.run(quantum_instance)

        x = sample_most_likely(result["eigvecs"][0])
        result["solution"] = max_cut.get_graph_solution(x)
        result["solution_objective"] = max_cut.max_cut_value(x, w)
        result["maxcut_objective"] = result["energy"] + offset
        """
        print("energy:", result["energy"])
        print("time:", result["eval_time"])
        print("max-cut objective:", result["energy"] + offset)
        print("solution:", max_cut.get_graph_solution(x))
        print("solution objective:", max_cut.max_cut_value(x, w))
        """
        return result


if __name__ == "__main__":
    unittest.main()



# This code is part of quantastica.qiskit_toaster
#
# (C) Copyright Quantastica 2019.
# https://quantastica.com/
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import uuid
from quantastica.qiskit_toaster import ToasterJob
from qiskit.providers import BaseBackend
from qiskit.providers.models import BackendConfiguration
import os

class ToasterBackend(BaseBackend):
    MAX_QUBITS_MEMORY = 32

    DEFAULT_CONFIGURATION = {'backend_name': 'toaster_simulator',
                             'backend_version': '1.0.0',
                             'n_qubits': MAX_QUBITS_MEMORY,
                             'url': 'https://quantastica.com/',
                             'simulator': True,
                             'local': True,
                             'conditional': False,
                             'open_pulse': False,
                             'memory': True,
                             'max_shots': 65536,
                             'description': 'An q-toaster based qasm simulator',
                             'coupling_map': None,
                             'basis_gates': ['u1',
                                             'u2',
                                             'u3',
                                             'cx',
                                             'id',
                                             'x',
                                             'y',
                                             'z',
                                             'h',
                                             's',
                                             't'],
                             'gates': []}

    def __init__(self, configuration=None,
                provider=None,
                backend_name = None,
                toasterpath=None):
        configuration = configuration or BackendConfiguration.from_dict(
            self.DEFAULT_CONFIGURATION)
        super().__init__(configuration=configuration, provider=provider)

        getstates = False
        if backend_name is not None and backend_name == "statevector_simulator":
            getstates = True

        self._getstates = getstates
        self._toasterpath = toasterpath or os.getenv('TOASTERPATH') or "qubit-toaster"

    #@profile
    def run(self, qobj):
        job_id = str(uuid.uuid4())
        job = ToasterJob.ToasterJob(self, job_id, qobj,
                                    self._toasterpath,
                                    getstates=self._getstates)
        job.submit()
        return job


    @staticmethod
    def name():
        return 'qubit_toaster'

def get_backend(backend_name = None, toasterpath = None):
        return ToasterBackend(backend_name = backend_name, toasterpath = toasterpath)
